//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <cassert>

#include <lib/chunk/list.h>
#include <lib/progname.h>
#include <lib/token/list.h>
#include <contrast/comparison_window.h>


comparison_window::~comparison_window()
{
}


comparison_window::comparison_window(const chunk_list &a_chunks,
        const token_list &a_lhs, const token_list &a_rhs) :
    chunks(a_chunks),
    lhs(a_lhs),
    rhs(a_rhs),
    side_by_side(true, 10),
    left_buttons(false, 10),
    format_left_as_left_button("Format as Self"),
    format_left_as_right_button("Format as Right"),
    right_buttons(false, 10),
    format_right_as_left_button("Format as Left"),
    format_right_as_right_button("Format as Self"),
    bottom_buttons(false, 10),
    back_button("Back"),
    next_button("Next"),
    quit_button(Gtk::Stock::QUIT),
    format_left_as(left_hand),
    format_right_as(right_hand),
    current_chunk(0)
{
    set_title(progname_get());
    set_border_width(5);
    set_default_size(1024, 768);

    left_text.property_editable() = false;
    right_text.property_editable() = false;

    add(vbox);
    vbox.pack_start(side_by_side);

    side_by_side.add(left_vbox);
    left_vbox.pack_start(left_scroller);
    left_scroller.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
    left_scroller.add(left_text);
    left_vbox.pack_start(left_buttons, Gtk::PACK_SHRINK);
    left_buttons.pack_start(format_left_as_left_button,
            Gtk::PACK_EXPAND_PADDING);
    left_buttons.pack_start(format_left_as_right_button,
            Gtk::PACK_EXPAND_PADDING);

    side_by_side.add(right_vbox);
    right_vbox.pack_start(right_scroller);
    right_scroller.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
    right_scroller.add(right_text);
    right_vbox.pack_start(right_buttons, Gtk::PACK_SHRINK);
    right_buttons.pack_start(format_right_as_left_button,
            Gtk::PACK_EXPAND_PADDING);
    right_buttons.pack_start(format_right_as_right_button,
            Gtk::PACK_EXPAND_PADDING);

    vbox.pack_start(bottom_buttons, Gtk::PACK_SHRINK);
    bottom_buttons.pack_start(back_button, Gtk::PACK_EXPAND_PADDING);
    bottom_buttons.pack_start(next_button, Gtk::PACK_EXPAND_PADDING);
    bottom_buttons.pack_start(quit_button, Gtk::PACK_EXPAND_PADDING);

    //Connect signals:
    quit_button.signal_clicked().connect(
            sigc::mem_fun(*this, &comparison_window::on_button_quit));
    format_left_as_left_button.signal_clicked().connect(
            sigc::mem_fun(*this, &comparison_window::left_as_left));
    format_left_as_right_button.signal_clicked().connect(
            sigc::mem_fun(*this, &comparison_window::left_as_right));
    format_right_as_left_button.signal_clicked().connect(
            sigc::mem_fun(*this, &comparison_window::right_as_left));
    format_right_as_right_button.signal_clicked().connect(
            sigc::mem_fun(*this, &comparison_window::right_as_right));
    back_button.signal_clicked().connect(
            sigc::mem_fun(*this, &comparison_window::go_back));
    next_button.signal_clicked().connect(
            sigc::mem_fun(*this, &comparison_window::go_next));

    left_red_background_tag = Gtk::TextBuffer::Tag::create();
    left_red_background_tag->property_background() = "#FFD0D0";
    left_text.get_buffer()->get_tag_table()->add(left_red_background_tag);

    left_red2_background_tag = Gtk::TextBuffer::Tag::create();
    left_red2_background_tag->property_background() = "#FF0000";
    left_text.get_buffer()->get_tag_table()->add(left_red2_background_tag);

    right_green_background_tag = Gtk::TextBuffer::Tag::create();
    right_green_background_tag->property_background() = "#D0FFD0";
    right_text.get_buffer()->get_tag_table()->add(right_green_background_tag);

    right_green2_background_tag = Gtk::TextBuffer::Tag::create();
    right_green2_background_tag->property_background() = "#00FF00";
    right_text.get_buffer()->get_tag_table()->add(right_green2_background_tag);

    left_redraw();
    right_redraw();

    show_all_children();
    go_first();
}


Glib::RefPtr<Gtk::TextBuffer::Mark>
buffer_mark_end(Glib::RefPtr<Gtk::TextBuffer> buffer, const rcstring &name)
{
    Glib::RefPtr<Gtk::TextBuffer::Mark> mark = buffer->get_mark(name.c_str());
    if (mark != 0)
        buffer->move_mark(mark, buffer->end());
    else
        mark = buffer->create_mark(name.c_str(), buffer->end(), true);
    return mark;
}


void
comparison_window::left_redraw()
{
    Glib::RefPtr<Gtk::TextBuffer> left_buffer = left_text.get_buffer();

    // Erase all text
    left_buffer->erase(left_buffer->begin(), left_buffer->end());

    for (size_t j = 0; j < chunks.size(); ++j)
    {
        rcstring left_mark_name = rcstring::printf("L%ld", long(j));
        Glib::RefPtr<Gtk::TextBuffer::Mark> left_mark =
            buffer_mark_end(left_buffer, left_mark_name);

        const chunk &c = chunks[j];
        for (size_t k = c.get_llo(); k < c.get_lhi(); ++k)
        {
            rcstring s =
                (
                    (
                        format_left_as == left_hand
                    ||
                        c.get_right_width() == 0
                    )
                ?
                    lhs[k].get_white()
                :
                    rhs[c.get_rlo() + k - c.get_llo()].get_white()
                );
            if (!s.empty())
            {
                left_buffer->insert_at_cursor(s.c_str(), s.c_str() + s.size());
            }
            s = lhs[k].get_ink();
            if (!s.empty())
            {
                left_buffer->insert_at_cursor(s.c_str(), s.c_str() + s.size());
            }
        }

        if (c.get_right_width() == 0)
        {
            left_buffer->apply_tag
            (
                left_red_background_tag,
                left_mark->get_iter(),
                left_buffer->end()
            );
        }
    }

    // One last mark at the end of the buffer.
    rcstring left_mark_name = rcstring::printf("L%ld", long(chunks.size()));
    Glib::RefPtr<Gtk::TextBuffer::Mark> left_mark =
        buffer_mark_end(left_buffer, left_mark_name);
}


void
comparison_window::right_redraw()
{
    Glib::RefPtr<Gtk::TextBuffer> right_buffer = right_text.get_buffer();

    // Erase all text
    right_buffer->erase(right_buffer->begin(), right_buffer->end());

    for (size_t j = 0; j < chunks.size(); ++j)
    {
        rcstring right_mark_name = rcstring::printf("R%ld", long(j));
        Glib::RefPtr<Gtk::TextBuffer::Mark> right_mark =
            buffer_mark_end(right_buffer, right_mark_name);

        const chunk &c = chunks[j];
        for (size_t k = c.get_rlo(); k < c.get_rhi(); ++k)
        {
            rcstring s =
                (
                    (
                        format_right_as == right_hand
                    ||
                        c.get_left_width() == 0
                    )
                ?
                    rhs[k].get_white()
                :
                    lhs[c.get_llo() + k - c.get_rlo()].get_white()
                );
            if (!s.empty())
            {
                right_buffer->insert_at_cursor(s.c_str(), s.c_str() + s.size());
            }
            s = rhs[k].get_ink();
            if (!s.empty())
            {
                right_buffer->insert_at_cursor(s.c_str(), s.c_str() + s.size());
            }
        }
        if (c.get_left_width() == 0)
        {
            right_buffer->apply_tag
            (
                right_green_background_tag,
                right_mark->get_iter(),
                right_buffer->end()
            );
        }
    }

    // One last mark at the end of the buffer.
    rcstring right_mark_name = rcstring::printf("R%ld", long(chunks.size()));
    Glib::RefPtr<Gtk::TextBuffer::Mark> right_mark =
        buffer_mark_end(right_buffer, right_mark_name);
}


void
comparison_window::on_button_quit()
{
    hide();
}


void
comparison_window::go_current()
{
    //
    // Special case for identical files.
    //
    Glib::RefPtr<Gtk::TextBuffer> left_buffer = left_text.get_buffer();
    Glib::RefPtr<Gtk::TextBuffer> right_buffer = right_text.get_buffer();
    if (chunks.size() == 1 && chunks[0].is_unchanged())
    {
        current_chunk = 0;
        left_buffer->place_cursor(left_buffer->begin());
        right_buffer->place_cursor(right_buffer->begin());
        return;
    }

    if (current_chunk >= chunks.size())
    {
        if (chunks.empty())
            current_chunk = 0;
        else
            current_chunk = chunks.size() - 1;
    }

    size_t end_chunk = current_chunk;
    if (end_chunk < chunks.size())
    {
        ++end_chunk;
        if
        (
            chunks[current_chunk].is_delete()
        &&
            end_chunk < chunks.size()
        &&
            chunks[end_chunk].is_insert()
        )
            ++end_chunk;
    }

    //
    // Clear the left highlight, if any.
    //
    left_buffer->remove_tag
    (
        left_red2_background_tag,
        left_buffer->begin(),
        left_buffer->end()
    );

    //
    // Find the start of the left highlighting.
    //
    rcstring left_begin_mark_name =
        rcstring::printf("L%ld", long(current_chunk));
    Glib::RefPtr<Gtk::TextBuffer::Mark> left_begin_mark =
        left_buffer->get_mark(left_begin_mark_name.c_str());
    assert(!!left_begin_mark);
    left_text.scroll_to(left_begin_mark, 0.05, 1.0, 0.5);
    Gtk::TextBuffer::iterator left_begin_iter = left_begin_mark->get_iter();

    //
    // Find the end of the left highlighting.
    //
    Gtk::TextBuffer::iterator left_end_iter = left_buffer->end();
    if (current_chunk < chunks.size())
    {
        rcstring left_end_mark_name =
            rcstring::printf("L%ld", long(end_chunk));
        Glib::RefPtr<Gtk::TextBuffer::Mark> left_end_mark =
            left_text.get_buffer()->get_mark(left_end_mark_name.c_str());
        assert(!!left_end_mark);

        left_end_iter = left_end_mark->get_iter();
    }

    //
    // Set the left hand highlighting.
    //
    left_buffer->apply_tag
    (
        left_red2_background_tag,
        left_begin_iter,
        left_end_iter
    );

    //
    // Clear the right highlight, if any.
    //
    right_buffer->remove_tag
    (
        right_green2_background_tag,
        right_buffer->begin(),
        right_buffer->end()
    );

    //
    // Find the start of the right highlighting.
    //
    rcstring right_begin_mark_name =
        rcstring::printf("R%ld", long(current_chunk));
    Glib::RefPtr<Gtk::TextBuffer::Mark> right_begin_mark =
        right_buffer->get_mark(right_begin_mark_name.c_str());
    assert(!!right_begin_mark);
    right_text.scroll_to(right_begin_mark, 0.05, 1.0, 0.5);
    Gtk::TextBuffer::iterator right_begin_iter = right_begin_mark->get_iter();

    //
    // Find the end of the right highlighting.
    //
    Gtk::TextBuffer::iterator right_end_iter = right_buffer->end();
    if (current_chunk < chunks.size())
    {
        rcstring right_end_mark_name =
            rcstring::printf("R%ld", long(end_chunk));
        Glib::RefPtr<Gtk::TextBuffer::Mark> right_end_mark =
            right_text.get_buffer()->get_mark(right_end_mark_name.c_str());
        assert(!!right_end_mark);

        right_end_iter = right_end_mark->get_iter();
    }

    //
    // Set the right hand highlighting.
    //
    right_buffer->apply_tag
    (
        right_green2_background_tag,
        right_begin_iter,
        right_end_iter
    );
}


void
comparison_window::go_back()
{
    if (current_chunk == 0)
        return;
    assert(chunks.size());
    size_t old_chunk = current_chunk;
    --current_chunk;
    while (current_chunk > 0 && chunks[current_chunk].is_unchanged())
        --current_chunk;

    //
    // If we are looking at the second half of a delete+insert pair,
    // more back to the delete portion.
    //
    if
    (
        chunks[current_chunk].is_insert()
    &&
        current_chunk > 0
    &&
        chunks[current_chunk - 1].is_delete()
    )
        --current_chunk;

    //
    // do nothing if this is the start of the files.
    //
    if (current_chunk == 0 && chunks.size() && chunks[0].is_unchanged())
    {
        current_chunk = old_chunk;
        return;
    }

    //
    // redraw with new selection.
    //
    go_current();
}


void
comparison_window::go_first()
{
    current_chunk = 0;
    while
    (
        current_chunk < chunks.size()
    &&
        chunks[current_chunk].is_unchanged()
    )
        ++current_chunk;

    //
    // Now move the highlighting.
    //
    go_current();
}


void
comparison_window::go_next()
{
    size_t old_chunk = current_chunk;

    //
    // We could be looking at a modification (left delete + right
    // insert), and a left delete , or a right insert.  But
    // modifications are represented by *two* chunks, because unchange
    // portions hand non-empty left and right halves.
    //
    if
    (
        chunks[current_chunk].is_delete()
    &&
        current_chunk + 1 < chunks.size()
    &&
        chunks[current_chunk + 1].is_insert()
    )
    {
        assert(chunks[current_chunk].get_lhi() ==
            chunks[current_chunk + 1].get_llo());
        assert(chunks[current_chunk].get_rhi() ==
            chunks[current_chunk + 1].get_rlo());
        ++current_chunk;
    }
    ++current_chunk;
    while
    (
        current_chunk < chunks.size()
    &&
        chunks[current_chunk].is_unchanged()
    )
        ++current_chunk;
    if (current_chunk >= chunks.size())
        current_chunk = old_chunk;

    //
    // Now move the highlighting.
    //
    go_current();
}


void
comparison_window::left_as_left()
{
    if (format_left_as == left_hand)
        return;
    format_left_as = left_hand;

    left_redraw();
    go_current();
}


void
comparison_window::left_as_right()
{
    if (format_left_as == right_hand)
        return;
    format_left_as = right_hand;

    left_redraw();
    go_current();
}


void
comparison_window::right_as_left()
{
    if (format_right_as == left_hand)
        return;
    format_right_as = left_hand;

    right_redraw();
    go_current();
}


void
comparison_window::right_as_right()
{
    if (format_right_as == right_hand)
        return;
    format_right_as = right_hand;

    right_redraw();
    go_current();
}
