//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <cstdio>
#include <cstdlib>
#include <unistd.h>
#include <gtkmm/main.h>

#include <lib/comparator.h>
#include <lib/is_regular_file.h>
#include <lib/print_version.h>
#include <lib/progname.h>
#include <lib/rcstring/accumulator.h>
#include <lib/source.h>
#include <lib/token/list.h>
#include <contrast/comparison_window.h>


static void
usage()
{
    const char *prog = progname_get();
    fprintf(stderr, "Usage: %s [ -<option>... ] <file1> <file2>\n", prog);
    fprintf(stderr, "       %s -V\n", prog);
    exit(1);
}


static rcstring
read_output(const rcstring &command)
{
    rcstring command_no_stderr = command + " 2> /dev/null";
    FILE *fp = popen(command_no_stderr.c_str(), "r");
    if (!fp)
        return rcstring();
    rcstring_accumulator ac;
    for (;;)
    {
        int c = getc(fp);
        if (c == EOF)
            break;
        if (c == '\n')
            break;
        ac.push_back((char)c);
    }
    fclose(fp);
    return ac.mkstr();
}


int
main(int argc, char **argv)
{
    const char *algorithm_name = 0;
    const char *tokenizer_name = 0;
    progname_set(argv[0]);
    Gtk::Main kit(argc, argv, false);
    for (;;)
    {
        int c = getopt(argc, argv, "a:lLt:V");
        if (c == EOF)
            break;
        switch (c)
        {
        case 'a':
            algorithm_name = optarg;
            break;

        case 'l':
            comparator::list();
            return 0;

        case 'L':
            source::list();
            return 0;

        case 't':
            tokenizer_name = optarg;
            break;

        case 'V':
            print_version();
            return 0;

        default:
            usage();
            // NOTREACHED
        }
    }
    token_list lhs;
    token_list rhs;
    switch (argc - optind)
    {
    default:
        usage();
        //NOTREACHED

    case 1:
        // look for Aegis
        {
            rcstring rhs_filename = argv[optind];
            rcstring temp = "${source " + rhs_filename.quote_shell() + " abs}";
            rcstring command = "aesub -bl " + temp.quote_shell();
            rcstring lhs_filename = read_output(command);
            if (!lhs_filename.empty())
            {
                lhs.read(lhs_filename, tokenizer_name);
                rhs.read(rhs_filename, tokenizer_name);
                break;
            }
        }

        // look for Subversion
        {
            rcstring rhs_filename = argv[optind];
            rcstring lhs_filename =
                ".svn/text-base/" + rhs_filename + ".svn-base";
            if (is_regular_file(lhs_filename))
            {
                lhs.read(lhs_filename, tokenizer_name);
                rhs.read(rhs_filename, tokenizer_name);
                break;
            }
        }

        // me no grok
        usage();
        //NOTREACHED

    case 2:
        lhs.read(argv[optind], tokenizer_name);
        rhs.read(argv[optind + 1], tokenizer_name);
        break;
    }

    comparator *cmp = comparator::factory(lhs, rhs, algorithm_name);
    cmp->compare();

    //
    // Show the window and return when it is closed.
    //
    comparison_window w(cmp->get_chunks(), lhs, rhs);
    kit.run(w);

    return 0;
}
