//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <cassert>
#include <cstdlib>

#include <lib/chunk/list.h>


chunk_list::~chunk_list()
{
    delete [] item;
    item = 0;
    length = 0;
    maximum = 0;
}


chunk_list::chunk_list() :
    item(0),
    length(0),
    maximum(0)
{
}


void
chunk_list::push_back(const chunk &arg)
{
    if (length >= maximum)
    {
        size_t new_maximum = maximum * 2 + 16;
        chunk *new_item = new chunk [new_maximum];
        for (size_t j = 0; j < length; ++j)
            new_item[j] = item[j];
        delete [] item;
        item = new_item;
        maximum = new_maximum;
    }
    item[length++] = arg;
}


static int
cmp(const void *va, const void *vb)
{
    const chunk &a = *(chunk *)va;
    const chunk &b = *(chunk *)vb;
    return (a == b ? 0 : (a < b ? -1 : 1));
}


void
chunk_list::sort()
{
    qsort(item, length, sizeof(item[0]), cmp);

    //
    // Now we merge all the chunks that we can.
    //
    for (;;)
    {
        chunk *out = item;
        chunk *end = item + length;
        chunk *in = item;
        while (in < end)
        {
            assert(in->valid());

            if (in[0].is_delete() && in + 1 < end && in[1].is_delete())
            {
                //
                // This is multiple deletes:
                //     { N1..N2, M..M }
                //     { N2..N3, M..M }
                // and it is replaced with:
                //     { N1..N3, M..M }
                //
                assert(in[0].get_lhi() == in[1].get_llo());
                assert(in[0].get_rlo() == in[1].get_rhi());
                *out++ =
                    chunk
                    (
                        in[0].get_llo(),
                        in[1].get_lhi(),
                        in[0].get_rlo(),
                        in[1].get_rhi()
                    );
                in += 2;
            }
            else if (in[0].is_insert() && in + 1 < end && in[1].is_insert())
            {
                //
                // This is multiple inserts:
                //     { N..N, M1..M2 }
                //     { N..N, M2..M3 }
                // and it is replacd with:
                //     { N..N, M1..M3 }
                //
                assert(in[0].get_llo() == in[1].get_lhi());
                assert(in[0].get_rhi() == in[1].get_rlo());
                *out++ =
                    chunk
                    (
                        in[0].get_llo(),
                        in[1].get_lhi(),
                        in[0].get_rlo(),
                        in[1].get_rhi()
                    );
                in += 2;
            }
            else if
            (
                in[0].is_unchanged()
            &&
                in + 1 < end
            &&
                in[1].is_unchanged()
            )
            {
                //
                // This is multiple unchanged:
                //     { N1..N2, M1..M2 }
                //     { N2..N3, M2..M3 }
                // it is replaced with:
                //     { N1..N3, M1..M3 }
                //
                assert(in[0].get_lhi() == in[1].get_llo());
                assert(in[0].get_rhi() == in[1].get_rlo());
                *out++ =
                    chunk
                    (
                        in[0].get_llo(),
                        in[1].get_lhi(),
                        in[0].get_rlo(),
                        in[1].get_rhi()
                    );
                in += 2;
            }
            else if
            (
                in + 4 <= end
            &&
                in[0].is_delete()
            &&
                in[1].is_insert()
            &&
                in[2].is_delete()
            &&
                in[3].is_insert()
            )
            {
                //
                // We could have mutiple ins/del sets:
                //     { N1..N2, M1..M1 }
                //     { N2..N2, M1..M2 }
                //     { N2..N3, M2..M2 }
                //     { N3..N3, M2..M3 }
                // replace with...
                //     { N1..N3, M1..M1 }
                //     { N3..N3, M1..M3 }
                //
                assert(in[0].get_lhi() == in[1].get_llo());
                assert(in[1].get_lhi() == in[2].get_llo());
                assert(in[2].get_lhi() == in[3].get_llo());
                assert(in[0].get_rhi() == in[1].get_rlo());
                assert(in[1].get_rhi() == in[2].get_rlo());
                assert(in[2].get_rhi() == in[3].get_rlo());
                size_t n1 = in[0].get_llo();
                size_t n3 = in[3].get_lhi();
                size_t m1 = in[0].get_rlo();
                size_t m3 = in[0].get_rhi();
                *out++ = chunk(n1, n3, m1, m1);
                *out++ = chunk(n3, n3, m1, m3);
                in += 4;
            }
            else
            {
                //
                // no change
                //
                *out++ = *in++;
            }
        }
        length = out - item;
        if (in == out)
            break;
    }
}


void
chunk_list::print()
    const
{
    for (size_t j = 0; j < length; ++j)
        item[j].print();
}
