//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <cassert>
#include <cstdio>

#include <lib/comparator.h>
#include <lib/token/list.h>


// #define DEBUG_COMPARATOR


#ifdef DEBUG_COMPARATOR
#define debug_print(fmt, args...) \
    fprintf(stderr, "%s: %d: " fmt "\n", __FILE__, __LINE__ , ## args )
#else
#define debug_print(fmt, args...)
#endif


comparator::~comparator()
{
    debug_print("%s", __PRETTY_FUNCTION__);
}


comparator::comparator(const token_list &a_lhs, const token_list &a_rhs) :
    lhs(a_lhs),
    rhs(a_rhs)
{
    debug_print("%s", __PRETTY_FUNCTION__);
}


void
comparator::compare()
{
    debug_print("%s", __PRETTY_FUNCTION__);
    //
    // Start with a region which spans both token sets,
    // and subdivide until there is nothing left.
    //
    subdivide(chunk(0, lhs.size(), 0, rhs.size()));

    //
    // We have to sort the chunk list because we jumbled it all up when
    // we were subdividing the sequences.  We did this on purpose to
    // reduce the stack depth.
    //
    chunks.sort();
}


//
// Recursive routine to find a minimal D-path through the edit graph of
// the two token strings.  Arguments are the beginning token numbers in
// the strings, and the number of tokens to examine.  This is basically
// a divide-and-conquer routine which finds the middle equal region of
// an optimal D-path, then calls itself to find the remainder of the
// path before the equal region and after the equal region.
//

void
comparator::subdivide(const chunk &look_here)
{
    debug_print("%s {", __PRETTY_FUNCTION__);
    chunk roi(look_here);
    for (;;)
    {
        debug_print("roi = %s", roi.representation().c_str());
        if (roi.get_left_width() == 0 || roi.get_right_width() == 0)
        {
            add_chunk(roi);
            return;
        }

        //
        // Find all of the common tokens at the start
        //
        size_t n;
        for (n = 0; n < roi.get_left_width() && n < roi.get_right_width(); ++n)
        {
            if (lhs[roi.get_llo() + n] != rhs[roi.get_rlo() + n])
                break;
        }

        //
        // Add the common start tokens to the chunk list.
        //
        if (n > 0)
        {
            add_chunk
            (
                chunk
                (
                    roi.get_llo(),
                    roi.get_llo() + n,
                    roi.get_rlo(),
                    roi.get_rlo() + n
                )
            );
            roi =
                chunk
                (
                    roi.get_llo() + n,
                    roi.get_lhi(),
                    roi.get_rlo() + n,
                    roi.get_rhi()
                );
            debug_print("roi = %s", roi.representation().c_str());
        }

        //
        // Find all the common tokens at the end.
        //
        for (n = 0; n < roi.get_left_width() && n < roi.get_right_width(); ++n)
        {
            if (lhs[roi.get_lhi() - 1 - n] != rhs[roi.get_rhi() - 1 - n])
                break;
        }

        //
        // Add the common end tokens to the chunk list.
        //
        if (n > 0)
        {
            size_t lpt = roi.get_lhi() - n;
            size_t rpt = roi.get_rhi() - n;
            add_chunk(chunk(lpt, roi.get_lhi(), rpt, roi.get_rhi()));
            roi = chunk(roi.get_llo(), lpt, roi.get_rlo(), rpt);
            debug_print("roi = %s", roi.representation().c_str());
        }

        if (roi.get_left_width() == 0 || roi.get_right_width() == 0)
        {
            add_chunk(roi);
            break;
        }

        //
        // Find a region on the left and the right which has equal tokens.
        //
        chunk eq;
        if (!find_equal_region(roi, eq))
        {
            //
            // If no common point is found, we have two separate chunks, one
            // for the deletes and one for the inserts.
            //
            add_chunk
            (
                chunk
                (
                    roi.get_llo(),
                    roi.get_lhi(),
                    roi.get_rlo(),
                    roi.get_rlo()
                )
            );
            add_chunk
            (
                chunk
                (
                    roi.get_lhi(),
                    roi.get_lhi(),
                    roi.get_rlo(),
                    roi.get_rhi()
                )
            );
            break;
        }

        //
        // An equal region was found.
        //
        debug_print("eq = %s", eq.representation().c_str());
        add_chunk(eq);

        //
        // Now we divide and conquer.  (But exploit tail recursion,
        // particularly for the dumber comparators.)
        //
        subdivide
        (
            chunk(roi.get_llo(), eq.get_llo(), roi.get_rlo(), eq.get_rlo())
        );
#if 0
        subdivide
        (
            chunk(eq.get_lhi(), roi.get_lhi(), eq.get_rhi(), roi.get_rhi())
        );
        break;
#else
        roi = chunk(eq.get_lhi(), roi.get_lhi(), eq.get_rhi(), roi.get_rhi());
#endif
    }
    debug_print("%s", "}");
}


void
comparator::add_chunk(const chunk &x)
{
    debug_print("%s", __PRETTY_FUNCTION__);
    debug_print("x = %s", x.representation().c_str());
    assert(x.valid());
#ifdef DEBUG_COMPARATOR
    for (size_t n = x.get_llo(); n < x.get_lhi(); ++n)
        debug_print("left[%d] = %s", (int)n, lhs[n].representation().c_str());
    for (size_t n = x.get_rlo(); n < x.get_rhi(); ++n)
        debug_print("right[%d] = %s", (int)n, rhs[n].representation().c_str());
#endif
    if (!x.empty())
        chunks.push_back(x);
}


void
comparator::print_chunks()
    const
{
    chunks.print();
}
