//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#ifndef LIB_COMPARATOR_H
#define LIB_COMPARATOR_H

#include <lib/chunk/list.h>
#include <lib/token/list.h>

/**
  * The abstract comparator base class is used to represent a token list
  * comparison engine.  No particular algorithm is assumed.
  */
class comparator
{
public:
    /**
      * The destructor.
      */
    virtual ~comparator();

    /**
      * The default constructor.
      *
      * @param lhs
      *     The left hand side of the comparison.  The value referenced
      *     MUST persist for longer than the life time of this object.
      * @param rhs
      *     The right hand side of the comparison.  The value referenced
      *     MUST persist for longer than the life time of this object.
      */
    comparator(const token_list &lhs, const token_list &rhs);

    /**
      * The compare method is used to compare the two token lists,
      * and produce a chunk list of the common and different parts.
      */
    void compare();

    /**
      * The factory class method is used to manufacture a comparator
      * instance based on the name of the algorithm.
      *
      *
      * @param lhs
      *     The left hand side of the comparison.
      * @param rhs
      *     The right hand side of the comparison.
      * @param algorithm_name
      *     The name of the comparison algorithm.
      * @returns
      *     a new comparator instance.  Use operator delete when you are
      *     done with it.
      */
    static comparator *factory(const token_list &lhs, const token_list &rhs,
        const rcstring &algorithm_name = "");

    static void list();

    void print_chunks() const;

    const chunk_list &get_chunks() const { return chunks; }

protected:
    /**
      * The find_equal_region method is used to locate a region within the
      * given region which has tokens on the left and the right agreeing.
      *
      * @param region
      *     The region in which to search for a common point.
      * @param equal_region
      *     The result, a sub-region with equal tokens.
      */
    virtual bool find_equal_region(const chunk &region, chunk &equal_region)
        = 0;

    //
    // The tokens_are_equal method is used to compare tokens from the
    // left and the right.  This is used by derived classes when they
    // are searching for a common point.
    //
    bool
    tokens_are_equal(size_t lmid, size_t rmid)
        const
    {
        return (lhs[lmid] == rhs[rmid]);
    }

protected:
    /**
      * The lhs instance variable is used to remember the tokens on the
      * left hand side of the comparison.
      */
    const token_list &lhs;

private:
    /**
      * The rhs instance variable is used to remember the tokens on the
      * right hand side of the comparison.
      */
    const token_list &rhs;

    /**
      * The chunks instance variable is used to remember the
      * subsequences of the left and right hand sides, showing the
      * commonalities and differences of the two token lists.
      */
    chunk_list chunks;

    /**
      * The subdivide method is used to break up the problem space into
      * smaller problem spaces.  Chunks are added to the chunk list as
      * they are discovered, nor necessarily in any particular order.
      *
      * @param region
      *     The region of interest
      */
    void subdivide(const chunk &region);

    void add_chunk(const chunk &region);

    /**
      * The default constructor.  Do not use.
      */
    comparator();

    /**
      * The copy constructor.  Do not use.
      */
    comparator(const comparator &);

    /**
      * The assignment operator.  Do not use.
      */
    comparator &operator=(const comparator &);
};

#endif // LIB_COMPARATOR_H
