//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009, 2010 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstdio>
#include <libexplain/output.h>

#include <lib/sizeof.h>
#include <lib/comparator/myers.h>
#include <lib/comparator/myers-eggert.h>


static comparator *
new_myers(const token_list &lhs, const token_list &rhs)
{
    return new comparator_myers(lhs, rhs);
}


static comparator *
new_myers_eggert(const token_list &lhs, const token_list &rhs)
{
    return new comparator_myers_eggert(lhs, rhs);
}


struct comparator_table_t
{
    rcstring name;
    comparator *(*func)(const token_list &lhs, const token_list &rhs);
    const char *description;
};


static comparator_table_t table[] =
{
    { "myers-eggert", new_myers_eggert,
            "Myers's algorithm with Paul Eggert's heuristic for large inputs" },
    { "meyers-eggert", new_myers_eggert, 0 },
    { "myers", new_myers, "Myers' O(ND) Difference Algorithm" },
    { "meyers", new_myers, 0 },
};


comparator *
comparator::factory(const token_list &lhs, const token_list &rhs,
    const rcstring &name)
{
    //
    // Make the name lower case for the comparisons.
    //
    rcstring lcname(name.empty() ? table[0].name : name.downcase());

    //
    // Scan the table for the name
    //
    for (comparator_table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (lcname == tp->name)
            return tp->func(lhs, rhs);
    }

    //
    // Oops.
    //
    explain_output_error_and_die("algorithm %s unknown", name.c_str());
    // NOTREACHED
    return 0;
}


void
comparator::list()
{
    printf("Available difference algorithms:\n");
    for (comparator_table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (!tp->description)
            continue;
        printf("%-15s %s", (tp->name + ":").c_str(), tp->description);
        if (tp == table)
            printf(" (default)");
        printf("\n");
    }
}
