//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//
// This file is derived from the src/analize.c file from GNU Diff.
//

#include <cassert>
#include <climits>

#include <lib/comparator/myers-eggert.h>


comparator_myers_eggert::~comparator_myers_eggert()
{
    delete [] fdiag_base;
    delete [] bdiag_base;
}


comparator_myers_eggert::comparator_myers_eggert(const token_list &a_lhs,
        const token_list &a_rhs) :
    comparator(a_lhs, a_rhs)
{
    //
    // Allocate vectors for the results of comparison: a flag for each
    // line of each file, saying whether that line is an insertion or
    // deletion.  Allocate an extra element, always 0, at each end of
    // each vector.
    //

    size_t diag_size = (a_lhs.size() + a_rhs.size() + 3);
    fdiag_base = new long [diag_size];
    bdiag_base = new long [diag_size];
    fdiag = fdiag_base + a_lhs.size() + 1;
    bdiag = bdiag_base + a_rhs.size() + 1;

    //
    // Set TOO_EXPENSIVE to be approximate square root of input size,
    // bounded below by 256.
    //
    too_expensive = 1;
    for (; diag_size != 0; diag_size >>= 2)
        too_expensive <<= 1;
    if (too_expensive < 256)
        too_expensive = 256;
    if (too_expensive > 10000)
        too_expensive = 10000;
}


//
// The basic algorithm is described in:
// "An O(ND) Difference Algorithm and its Variations", Eugene Myers,
// Algorithmica Vol. 1 No. 2, 1986, pp. 251-266; see especially section
// 4.2, which describes the variation used below.
//
// This code uses the TOO_EXPENSIVE heuristic, by Paul Eggert, to limit
// the cost to O(N**1.5 log N) at the price of producing suboptimal
// output for large inputs with many differences.
//
// The basic algorithm was independently discovered as described in:
// "Algorithms for Approximate String Matching", E. Ukkonen, Information
// and Control Vol. 64, 1985, pp. 100-118.
//

/**
  * Find the midpoint of the shortest edit script for a specified
  * portion of the two files.
  *
  * Scan from the beginnings of the files, and simultaneously from the ends,
  * doing a breadth-first search through the space of edit-sequence.
  * When the two searches meet, we have found the midpoint of the shortest
  * edit sequence.
  *
  * If FIND_MINIMAL is nonzero, find the minimal edit script regardless
  * of expense.  Otherwise, if the search is too expensive, use
  * heuristics to stop the search and report a suboptimal answer.
  *
  * Set PART->(xmid,ymid) to the midpoint (XMID,YMID).  The diagonal number
  * XMID - YMID equals the number of inserted lines minus the number
  * of deleted lines (counting only lines before the midpoint).
  * Return the approximate edit cost; this is the total number of
  * lines inserted or deleted (counting only lines before the midpoint),
  * unless a heuristic is used to terminate the search prematurely.
  *
  * Set PART->lo_minimal to true iff the minimal edit script for the
  * left half of the partition is known; similarly for PART->hi_minimal.
  *
  * This function assumes that the first lines of the specified portions
  * of the two files do not match, and likewise that the last lines do not
  * match.  The caller must trim matching lines from the beginning and end
  * of the portions it is going to specify.
  *
  * If we return the "wrong" partitions,
  * the worst this can do is cause suboptimal diff output.
  * It cannot cause incorrect diff output.
  */
bool
comparator_myers_eggert::find_equal_region(const chunk &region,
    chunk &equal_region)
{
    long xoff = region.get_llo();
    long xlim = region.get_lhi();
    long yoff = region.get_rlo();
    long ylim = region.get_rhi();
    assert(xoff < xlim);
    assert(yoff < ylim);
    if (xoff == xlim || yoff == ylim)
        return false;

    long const dmin = xoff - ylim;      // Minimum valid diagonal.
    long const dmax = xlim - yoff;      // Maximum valid diagonal.
    long const fmid = xoff - yoff;      // Center diagonal of top-down search.
    long const bmid = xlim - ylim;      // Center diagonal of bottom-up search.
    long fmin = fmid;
    long fmax = fmid;   // Limits of top-down search.
    long bmin = bmid;
    long bmax = bmid;   // Limits of bottom-up search.

    //
    // True if southeast corner is on an odd diagonal with respect to
    // the northwest.
    //
    bool odd = (((fmid - bmid) & 1) != 0);

    fdiag[fmid] = xoff;
    bdiag[bmid] = xlim;

    for (long c = 1;; ++c)
    {
        bool big_snake = false;

        // Extend the top-down search by an edit step in each diagonal.
        if (fmin > dmin)
            fdiag[--fmin - 1] = -1;
        else
            ++fmin;
        if (fmax < dmax)
            fdiag[++fmax + 1] = -1;
        else
            --fmax;
        for (long d = fmax; d >= fmin; d -= 2)
        {
            long tlo = fdiag[d - 1];
            long thi = fdiag[d + 1];
            long x = 0;
            if (tlo >= thi)
                x = tlo + 1;
            else
                x = thi;
            long oldx = x;
            long y = x - d;
            while (x < xlim && y < ylim && tokens_are_equal(x, y))
            {
                ++x;
                ++y;
            }
            if (x - oldx > SNAKE_LIMIT)
                big_snake = true;
            fdiag[d] = x;
            if (odd && bmin <= d && d <= bmax && bdiag[d] <= x)
            {
                equal_region = chunk(x, x, y, y);
                return true;
            }
        }

        // Similarly extend the bottom-up search.
        if (bmin > dmin)
            bdiag[--bmin - 1] = LONG_MAX;
        else
            ++bmin;
        if (bmax < dmax)
            bdiag[++bmax + 1] = LONG_MAX;
        else
            --bmax;
        for (long d = bmax; d >= bmin; d -= 2)
        {
            long tlo = bdiag[d - 1];
            long thi = bdiag[d + 1];
            long x = 0;
            if (tlo < thi)
                x = tlo;
            else
                x = thi - 1;
            long oldx = x;
            long y = x - d;
            while (x > xoff && y > yoff && tokens_are_equal(x - 1, y - 1))
            {
                --x;
                --y;
            }
            if (oldx - x > SNAKE_LIMIT)
                big_snake = true;
            bdiag[d] = x;
            if (!odd && fmin <= d && d <= fmax && x <= fdiag[d])
            {
                equal_region = chunk(x, x, y, y);
                return true;
            }
        }

        //
        // if we were looking for the minimal solution
        //    continue;
        // but we aren't.  If you want that, use the other comparator.
        //

        //
        // Heuristic: check occasionally for a diagonal that has made
        // lots of progress compared with the edit distance.
        // If we have any such, find the one that has made the most
        // progress and return it as if it had succeeded.
        //
        // With this heuristic, for files with a constant small density
        // of changes, the algorithm is linear in the file size.
        //
        if (200 < c && big_snake)
        {
            long best = 0;
            for (long d = fmax; d >= fmin; d -= 2)
            {
                long dd = d - fmid;
                long x = fdiag[d];
                long y = x - d;
                long v = (x - xoff) * 2 - dd;
                if (v > 12 * (c + (dd < 0 ? -dd : dd)))
                {
                    if
                    (
                        v > best
                    &&
                        xoff + SNAKE_LIMIT <= x
                    &&
                        x < xlim
                    &&
                        yoff + SNAKE_LIMIT <= y
                    &&
                        y < ylim
                    )
                    {
                        //
                        // We have a good enough best diagonal;
                        // now insist that it end with a significant snake.
                        //
                        for (int k = 1; tokens_are_equal(x - k, y - k); k++)
                        {
                            if (k == SNAKE_LIMIT)
                            {
                                best = v;
                                equal_region = chunk(x, x, y, y);
                                break;
                            }
                        }
                    }
                }
            }
            if (best > 0)
                return true;

            best = 0;
            for (long d = bmax; d >= bmin; d -= 2)
            {
                long dd = d - bmid;
                long x = bdiag[d];
                long y = x - d;
                long v = (xlim - x) * 2 + dd;
                if (v > 12 * (c + (dd < 0 ? -dd : dd)))
                {
                    if
                    (
                        v > best
                    &&
                        xoff < x
                    &&
                        x <= xlim - SNAKE_LIMIT
                    &&
                        yoff < y
                    &&
                        y <= ylim - SNAKE_LIMIT
                    )
                    {
                        //
                        // We have a good enough best diagonal;
                        // now insist that it end with a significant snake.
                        //
                        for (int k = 0; tokens_are_equal(x + k, y + k); k++)
                        {
                            if (k == SNAKE_LIMIT - 1)
                            {
                                equal_region = chunk(x, x, y, y);
                            }
                        }
                    }
                }
            }
            if (best > 0)
            {
                return true;
            }
        }

        //
        // Heuristic: if we've gone well beyond the call of duty,
        // give up and report halfway between our best results so far.
        //
        if (c >= too_expensive)
        {
            long fxbest = 0;
            long bxbest = 0;

            // Find forward diagonal that maximizes X + Y.
            long fxybest = -1;
            for (long d = fmax; d >= fmin; d -= 2)
            {
                long x = (fdiag[d] < xlim ? fdiag[d] : xlim);
                long y = x - d;
                if (ylim < y)
                    x = ylim + d, y = ylim;
                if (fxybest < x + y)
                {
                    fxybest = x + y;
                    fxbest = x;
                }
            }

            // Find backward diagonal that minimizes X + Y.
            long bxybest = LONG_MAX;
            for (long d = bmax; d >= bmin; d -= 2)
            {
                long x = (xoff > bdiag[d] ? xoff : bdiag[d]);
                long y = x - d;
                if (y < yoff)
                    x = yoff + d, y = yoff;
                if (x + y < bxybest)
                {
                    bxybest = x + y;
                    bxbest = x;
                }
            }

            // Use the better of the two diagonals.
            if ((xlim + ylim) - bxybest < fxybest - (xoff + yoff))
            {
                long x = fxbest;
                long y = fxybest - fxbest;
                equal_region = chunk(x, x, y, y);
            }
            else
            {
                long x = bxbest;
                long y = bxybest - bxbest;
                equal_region = chunk(x, x, y, y);
            }
            return true;
        }
    }
}
