//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <cassert>
#include <cstdio>

#include <lib/comparator/myers.h>


// #define DEBUG_COMPARATOR_MYERS

#ifdef DEBUG_COMPARATOR_MYERS
#define debug_print(fmt, args...) \
        fprintf(stderr, "%s: %d: " fmt "\n", __FILE__, __LINE__ , ## args)
#else
#define debug_print(fmt, args...) \
        (void)0
#endif



comparator_myers::~comparator_myers()
{
    debug_print("%s", __PRETTY_FUNCTION__);
    delete [] v1_table;
    delete [] v2_table;
}


comparator_myers::comparator_myers(const token_list &a_lhs,
        const token_list &a_rhs) :
    comparator(a_lhs, a_rhs),
    v1_table(0),
    v2_table(0),
    v1(0),
    v2(0)
{
    debug_print("%s", __PRETTY_FUNCTION__);
    size_t n = a_lhs.size();
    size_t m = a_rhs.size();
    size_t maxtoks = (n > m ? n : m);

    size_t tablesize = maxtoks * 2 + 1;
    v1_table = new long [tablesize];
    v2_table = new long [tablesize];
    v1 = v1_table + maxtoks;
    v2 = v2_table + maxtoks;
}


//
// Routine to find the middle equal region of an optimial D-path
// spanning lines A to A+N in file A to lines B to B+N in file B.
// Returns the length D of the D-path as a return value, and the upper
// left and lower right relative coordinates of a equal region midway
// through the D-path.
//

bool
comparator_myers::find_equal_region(const chunk &region, chunk &equal_region)
{
    debug_print("%s {", __PRETTY_FUNCTION__);
    debug_print("region = %s", region.representation().c_str());

    long DELTA = region.get_left_width() - region.get_right_width();
    int odd = DELTA & 1;
    long MAXD = (region.get_right_width() + region.get_left_width() + 1) / 2;
    v1[1] = 0;
    v2[-1] = 0;

    //
    // This is the main loop for searching for the equal region.
    // D is the distance off the diagonals, and is the number
    // of changes needed to get from the upper left to the
    // lower right corner of the region.
    //
    for (long D = 0; D <= MAXD; D++)
    {
        debug_print("D = %ld", D);
        //
        // Examine all diagonals within current distance.
        // First search from upper left to lower right,
        // and then search from lower right to upper left.
        //
        debug_print("forward");
        for (long k = -D; k <= D; k += 2)
        {
            debug_print("k = %ld", k);
            //
            // Find the end of the furthest forward D-path
            // in diagonal k.
            //
            long x = 0;
            if (k == -D || (k != D && (v1[k - 1] < v1[k + 1])))
                x = v1[k + 1];
            else
                x = v1[k - 1] + 1;
            long y = x - k;
            long oldx = x;
            while
            (
                x < long(region.get_left_width())
            &&
                y < long(region.get_right_width())
            &&
                tokens_are_equal(region.get_llo() + x, region.get_rlo() + y)
            )
            {
                debug_print("x = %ld, y = %ld, token = %s", x, y,
                    lhs[region.get_llo() + x].representation().c_str());
                x++;
                y++;
            }
            v1[k] = x;

            //
            // See if path overlaps furthest reverse D-path.
            // If so, then we have found the equal region.
            //
            if (odd && (k >= (DELTA - (D - 1))) && (k <= (DELTA + (D - 1))))
            {
                if ((x + v2[k - DELTA]) >= long(region.get_left_width()))
                {
                    debug_print("found");
                    long ello = region.get_llo() + oldx;
                    long elhi = region.get_llo() + x;
                    long erlo = region.get_rlo() + oldx - k;
                    long erhi = region.get_rlo() + y;
                    debug_print("ello = %ld", ello);
                    debug_print("elhi = %ld", elhi);
                    debug_print("erlo = %ld", erlo);
                    debug_print("erhi = %ld", erhi);
                    equal_region = chunk(ello, elhi, erlo, erhi);
                    return true;
                }
            }
        }

        debug_print("reverse");
        for (long k = -D; k <= D; k += 2)
        {
            debug_print("k = %ld", k);
            //
            // Find the end of the furthest reaching reverse
            // path in diagonal k+DELTA.
            //
            long x = 0;
            if (k == D || (k != -D && (v2[k + 1] < v2[k - 1])))
                x = v2[k - 1];
            else
                x = v2[k + 1] + 1;
            long y = x + k;
            long oldx = x;
            while
            (
                x < long(region.get_left_width())
            &&
                y < long(region.get_right_width())
            &&
                tokens_are_equal
                (
                    region.get_lhi() - 1 - x,
                    region.get_rhi() - 1 - y
                )
            )
            {
                debug_print("x = %ld, y = %ld, token = %s", x, y,
                    lhs[region.get_lhi() - 1 - x].representation().c_str());
                x++;
                y++;
            }
            v2[k] = x;

            //
            // See if path overlaps furthest forward D-path.
            // If so, then we have found the equal region.
            //
            if (!odd && (k <= D - DELTA) && (k >= -D - DELTA))
            {
                if ((x + v1[k + DELTA]) >= long(region.get_left_width()))
                {
                    debug_print("found");
                    long ello = region.get_lhi() - x;
                    long elhi = region.get_lhi() - oldx;
                    long erlo = region.get_rhi() - y;
                    long erhi = erlo + (elhi - ello);
                    debug_print("ello = %ld", ello);
                    debug_print("elhi = %ld", elhi);
                    debug_print("erlo = %ld", erlo);
                    debug_print("erhi = %ld", erhi);
                    equal_region = chunk(ello, elhi, erlo, erhi);
                    return true;
                }
            }
        }
    }

    //
    // Middle equal region procedure failed!
    //
    assert(false);
    debug_print("}");
    return false;
}
