//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#ifndef LIB_SOURCE_H
#define LIB_SOURCE_H

#include <lib/rcstring.h>

/**
  * The source class is used to represent a source of tokens.
  * Specializations of this interface provide for language specific
  * tokenization.
  */
class source
{
public:
    /**
      * The destructor.
      */
    virtual ~source();

    /**
      * The factory class method is used to create new tokenizers.
      *
      * @param filename
      *     The name of the file which is the source of the tokens.
      * @param tokenizer_name
      *     The name of the tokenizer (class) to be used.
      * @returns
      *     new source, use operator delete when you are done with it.
      */
    static source *factory(const rcstring &filename,
        const rcstring &tokenizer_name);

    /**
      * The list method is used to list the tokenizers known to the
      * factory class method.  The names and descriptions are written to
      * the standard output.
      */
    static void list();

protected:
    /**
      * The default constructor.
      * Only derived classes may access.
      */
    source(const rcstring &filename);

public:
    enum type_t
    {
        type_whitespace,
        type_ink,
        type_end_of_file
    };

    /**
      * The get_next_token method is used to obtain the next token from
      * the input source.
      */
    virtual type_t get_next_token(rcstring &value) = 0;

protected:
    /**
      * The getch method is used to obtain the next character from the
      * input file.
      *
      * @returns
      *     -1 at end-of-file, positive numbers for printable characters
      */
    int
    getch()
    {
        if (pushback_length > 0)
            return pushback[--pushback_length];
        if (buffer_pos < buffer_end)
            return *buffer_pos++;
        return underflow();
    }

    /**
      * The ungetch method is used to return a character to the input,
      * so that it may be used for the following token.
      *
      * @param c
      *     a character previously returned by the getch method.
      */
    void ungetch(int c);

private:
    /**
      * The filename instance variable is used to remember the name of
      * the file being read.
      */
    rcstring filename;

    /**
      * The line_number instance variable is used to remember the
      * current line number of the input file being read.
      */
    int line_number;

    /**
      * The fd instance variable is used to remember the file descriptor
      * of the file open for reading.
      */
    int fd;

    /**
      * The buffer instance variable is used to remember the base of a
      * dynamically allocated array used to buffer the input.
      */
    unsigned char *buffer;

    /**
      * The buffer_size instance variable is used to remember how much
      * space was allocated to the buffer instance variable.
      */
    size_t buffer_size;

    /**
      * The buffer_pos instance variable is used to remember the next
      * character to be read from the buffer.
      */
    unsigned char *buffer_pos;

    /**
      * The buffer_end instance variable is used to remember the end of
      * the data read into the buffer.
      */
    unsigned char *buffer_end;

    /**
      * The pushback instance variable is used to remember the
      * characters given to ungetch.
      */
    unsigned char pushback[4];

    /**
      * The pushback_length instance variable is used to remember how
      * many characters are in the push back buffer.
      */
    size_t pushback_length;

    /**
      * The underflow method is used by the getch method to obtain the
      * next character, when the current buffer is exhausted.
      */
    int underflow();

    /**
      * The default constructor.  Do not use.
      */
    source();

    /**
      * The copy constructor.  Do not use.
      */
    source(const source &);

    /**
      * The assignment operator.  Do not use.
      */
    source &operator=(const source &);
};

#endif // LIB_SOURCE_H
