//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <cassert>
#include <cctype>
#include <cstdio>

#include <lib/source/c.h>


// #define DEBUG_TOKENIZER


#ifdef DEBUG_TOKENIZER
#define debug_print(fmt, args...) \
    fprintf(stderr, "%s: %d: " fmt "\n", __FILE__, __LINE__ , ## args )
#else
#define debug_print(fmt, args...)
#endif


source_c::~source_c()
{
    debug_print("%s", __PRETTY_FUNCTION__);
}


source_c::source_c(const rcstring &a_filename) :
    source(a_filename),
    state(state_normal)
{
    debug_print("%s", __PRETTY_FUNCTION__);
}


source::type_t
source_c::get_next_token(rcstring &value)
{
    debug_print("%s", __PRETTY_FUNCTION__);
    ac.clear();
    switch (state)
    {
    case state_normal:
        return get_normal_token(value);

    case state_c_comment:
        return get_c_comment_token(value);

    case state_cpp_comment:
        return get_cpp_comment_token(value);

    case state_string_constant:
        return get_string_token(value, '"');

    case state_char_constant:
        return get_string_token(value, '\'');
    }
    assert(false);
    return type_end_of_file;
}


source::type_t
source_c::get_normal_token(rcstring &value)
{
    unsigned char uc2;
    unsigned char uc3;
    debug_print("%s", __PRETTY_FUNCTION__);
    int c = getch();
    if (c < 0)
    {
        value = "";
        return type_end_of_file;
    }
    unsigned char uc = c;
    debug_print("uc = %s", rcstring::printf("%c", uc).quote_c().c_str());
    switch (uc)
    {
    case ' ':
    case '\b':
    case '\f':
    case '\n':
    case '\r':
    case '\t':
    case '\v':
        for (;;)
        {
            ac.push_back(uc);
            c = getch();
            if (c < 0)
                break;
            uc = c;
            switch (uc)
            {
            case ' ':
            case '\b':
            case '\f':
            case '\n':
            case '\r':
            case '\t':
            case '\v':
                continue;
            }
            ungetch(uc);
            break;
        }
        value = ac.mkstr();
        return type_whitespace;

    case '!':
    case '%':
    case '*':
    case '^':
        ac.push_back(uc);
        c = getch();
        if (c < 0)
            break;
        uc2 = c;
        if (uc2 == '=')
            ac.push_back(uc2);
        else
            ungetch(uc2);
        break;

    case '#':
    case '=':
    case ':':
        ac.push_back(uc);
        c = getch();
        if (c < 0)
            break;
        uc2 = c;
        if (uc == uc2)
            ac.push_back(uc2);
        else
            ungetch(uc2);
        break;

    case '&':
    case '+':
    case '|':
        ac.push_back(uc);
        c = getch();
        if (c < 0)
            break;
        uc2 = c;
        if (uc2 == uc || uc2 == '=')
            ac.push_back(uc2);
        else
            ungetch(uc2);
        break;

    case '-':
        ac.push_back(uc);
        c = getch();
        if (c < 0)
            break;
        uc2 = c;
        if (uc2 == '-' || uc2 == '=' || uc2 == '>')
            ac.push_back(uc2);
        else
            ungetch(uc2);
        break;

    case '.':
        ac.push_back(uc);
        c = getch();
        if (c < 0)
            break;
        uc2 = c;
        if (uc2 != '.')
        {
            if (isdigit(uc2))
                goto number;
            ungetch(uc2);
            break;
        }
        c = getch();
        if (c < 0)
        {
            ungetch('.');
            break;
        }
        uc3 = c;
        if (uc3 != '.')
        {
            ungetch(uc3);
            ungetch(uc2);
            break;
        }
        ac.push_back(uc2);
        ac.push_back(uc3);
        break;

    case '<':
    case '>':
        ac.push_back(uc);
        c = getch();
        if (c < 0)
            break;
        uc2 = c;
        if (uc2 == uc)
        {
            ac.push_back(uc2);
            c = getch();
            if (c < 0)
                break;
            uc2 = c;
        }
        if (uc2 == '=')
            ac.push_back(uc2);
        else
            ungetch(uc2);
        break;

    case '/':
        ac.push_back(uc);
        c = getch();
        if (c < 0)
            break;
        uc2 = c;
        switch (uc2)
        {
        case '/':
            ac.push_back(uc2);
            debug_print("state = C++ comment");
            state = state_cpp_comment;
            break;

        case '*':
            ac.push_back(uc2);
            debug_print("state = C comment");
            state = state_c_comment;
            break;

        case '=':
            ac.push_back(uc2);
            break;

        default:
            ungetch(uc2);
            break;
        }
        break;

    case '\'':
        debug_print("state = 'char' constant");
        state = state_char_constant;
        value = "'";
        return type_ink;

    case '"':
        debug_print("state = \"string\" constant");
        state = state_string_constant;
        value = "\"";
        return type_ink;

    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
        number:
        for (;;)
        {
            ac.push_back(uc);

            switch (uc)
            {
            case 'E':
            case 'P':
            case 'e':
            case 'p':
                //
                // This can be part of floating point mantissa,
                // for example 1e-6 or 0x1p-6
                //
                c = getch();
                if (c == '+' || c == '-')
                {
                    uc = c;
                    ac.push_back(uc);
                }
                else
                    ungetch(c);
                break;
            }

            c = getch();
            if (c < 0)
                break;
            uc = c;
            switch (uc)
            {
            case '.':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
                continue;

            case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
            case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
            case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
            case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
            case 'Y': case 'Z':
            case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
            case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
            case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
            case 's': case 't': case 'u': case 'v': case 'w': case 'x':
            case 'y': case 'z':
                // In order to support token glueing, and floating point
                // exponents, the range of junk allowed in a numeric
                // preprocessing token is amazing.
                // It isn't just [eExX] as you may first expect.
                continue;

            default:
                ungetch(uc);
                break;
            }
            break;
        }
        break;

    case '$':
    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F': case 'G':
    case 'H': case 'I': case 'J': case 'K': case 'L': case 'M': case 'N':
    case 'O': case 'P': case 'Q': case 'R': case 'S': case 'T': case 'U':
    case 'V': case 'W': case 'X': case 'Y': case 'Z':
    case '_':
    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f': case 'g':
    case 'h': case 'i': case 'j': case 'k': case 'l': case 'm': case 'n':
    case 'o': case 'p': case 'q': case 'r': case 's': case 't': case 'u':
    case 'v': case 'w': case 'x': case 'y': case 'z':
        for (;;)
        {
            ac.push_back(uc);

            c = getch();
            if (c < 0)
                break;
            uc = c;
            switch (uc)
            {
            case '$':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
            case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
            case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
            case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
            case 'Y': case 'Z':
            case '_':
            case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
            case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
            case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
            case 's': case 't': case 'u': case 'v': case 'w': case 'x':
            case 'y': case 'z':
                continue;

            default:
                ungetch(uc);
                break;
            }
            break;
        }
        break;

    default:
        // Everything else is a single character symbol
        ac.push_back(uc);
        break;
    }
    value = ac.mkstr();
    return type_ink;
}


source::type_t
source_c::get_cpp_comment_token(rcstring &value)
{
    debug_print("%s", __PRETTY_FUNCTION__);
    int c = getch();
    if (c < 0)
    {
        value = "";
        return type_end_of_file;
    }
    unsigned char uc = c;
    debug_print("c = %s", rcstring::printf("%c", uc).quote_c().c_str());
    switch (uc)
    {
    case '\n':
        debug_print("state = normal");
        state = state_normal;
        // fall through...

    case ' ':
    case '\b':
    case '\f':
    case '\r':
    case '\t':
    case '\v':
        for (;;)
        {
            ac.push_back(uc);

            c = getch();
            if (c < 0)
                break;
            uc = c;
            switch (uc)
            {
            case '\n':
                debug_print("state = normal");
                state = state_normal;
                continue;

            case ' ':
            case '\b':
            case '\f':
            case '\r':
            case '\t':
            case '\v':
                continue;
            }
            ungetch(uc);
            break;
        }
        value = ac.mkstr();
        return type_whitespace;

    case '$':
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
    case 'A': case 'B': case 'C': case 'D': case 'E':
    case 'F': case 'G': case 'H': case 'I': case 'J':
    case 'K': case 'L': case 'M': case 'N': case 'O':
    case 'P': case 'Q': case 'R': case 'S': case 'T':
    case 'U': case 'V': case 'W': case 'X': case 'Y':
    case 'Z':
    case '_':
    case 'a': case 'b': case 'c': case 'd': case 'e':
    case 'f': case 'g': case 'h': case 'i': case 'j':
    case 'k': case 'l': case 'm': case 'n': case 'o':
    case 'p': case 'q': case 'r': case 's': case 't':
    case 'u': case 'v': case 'w': case 'x': case 'y':
    case 'z':
        for (;;)
        {
            ac.push_back(uc);

            //
            // Get the next character
            //
            c = getch();
            if (c < 0)
            {
                debug_print("state = normal");
                state = state_normal;
                break;
            }
            uc = c;
            switch (uc)
            {
            case '$':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y':
            case 'Z':
            case '_':
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case 'w': case 'x': case 'y':
            case 'z':
                continue;

            default:
                ungetch(uc);
                break;
            }
            break;
        }
        value = ac.mkstr();
        return type_ink;

    default:
        for (;;)
        {
            ac.push_back(uc);

            c = getch();
            if (c < 0)
                break;
            uc = c;
            switch (uc)
            {
            case ' ':
            case '\b':
            case '\f':
            case '\n':
            case '\r':
            case '\t':
            case '\v':
            case '$':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y':
            case 'Z':
            case '_':
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case 'w': case 'x': case 'y':
            case 'z':
                ungetch(uc);
                break;

            default:
                continue;
            }
            break;
        }
        value = ac.mkstr();
        return type_ink;
    }
}


source::type_t
source_c::get_c_comment_token(rcstring &value)
{
    debug_print("%s", __PRETTY_FUNCTION__);
    int c = getch();
    if (c < 0)
    {
        value = "";
        debug_print("state = normal");
        state = state_normal;
        return type_end_of_file;
    }
    unsigned char uc = c;
    debug_print("c = %s", rcstring::printf("%c", uc).quote_c().c_str());
    switch (uc)
    {
    case ' ':
    case '\b':
    case '\f':
    case '\n':
    case '\r':
    case '\t':
    case '\v':
        for (;;)
        {
            ac.push_back(uc);
            c = getch();
            if (c < 0)
                break;
            uc = c;
            switch (uc)
            {
            case ' ':
            case '\b':
            case '\f':
            case '\n':
            case '\r':
            case '\t':
            case '\v':
                continue;
            }
            ungetch(uc);
            break;
        }
        value = ac.mkstr();
        return type_whitespace;

    case '$':
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
    case 'A': case 'B': case 'C': case 'D': case 'E':
    case 'F': case 'G': case 'H': case 'I': case 'J':
    case 'K': case 'L': case 'M': case 'N': case 'O':
    case 'P': case 'Q': case 'R': case 'S': case 'T':
    case 'U': case 'V': case 'W': case 'X': case 'Y':
    case 'Z':
    case '_':
    case 'a': case 'b': case 'c': case 'd': case 'e':
    case 'f': case 'g': case 'h': case 'i': case 'j':
    case 'k': case 'l': case 'm': case 'n': case 'o':
    case 'p': case 'q': case 'r': case 's': case 't':
    case 'u': case 'v': case 'w': case 'x': case 'y':
    case 'z':
        for (;;)
        {
            ac.push_back(uc);

            //
            // Get the next character
            //
            c = getch();
            if (c < 0)
            {
                debug_print("state = normal");
                state = state_normal;
                break;
            }
            uc = c;
            switch (uc)
            {
            case '$':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y':
            case 'Z':
            case '_':
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case 'w': case 'x': case 'y':
            case 'z':
                continue;

            default:
                ungetch(uc);
                break;
            }
            break;
        }
        value = ac.mkstr();
        return type_ink;

    case '*':
        c = getch();
        if (c < 0)
            break;
        uc = c;
        if (c == '/')
        {
            value = "*/";
            debug_print("state = normal");
            state = state_normal;
            return type_ink;
        }
        ungetch(uc);
        c = '*';
        uc = c;
        // Fall through...

    default:
        for (;;)
        {
            ac.push_back(uc);

            c = getch();
            if (c < 0)
                break;
            uc = c;
            switch (uc)
            {
            case ' ':
            case '\b':
            case '\f':
            case '\n':
            case '\r':
            case '\t':
            case '\v':
            case '$':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y':
            case 'Z':
            case '_':
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case 'w': case 'x': case 'y':
            case 'z':
                ungetch(uc);
                break;

            case '*':
                c = getch();
                ungetch(c);
                if (c != '/')
                    continue;
                ungetch('*');
                break;

            default:
                continue;
            }
            break;
        }
    }
    value = ac.mkstr();
    return type_ink;
}


source::type_t
source_c::get_string_token(rcstring &value, unsigned char terminator)
{
    debug_print("%s", __PRETTY_FUNCTION__);
    int c = getch();
    if (c < 0)
    {
        value = "";
        return type_end_of_file;
    }
    unsigned char uc = c;
    debug_print("uc = %s", rcstring::printf("%c", uc).quote_c().c_str());
    switch (uc)
    {
    case '\n':
        debug_print("state = normal");
        state = state_normal;
        // fall through...

    case ' ':
    case '\b':
    case '\f':
    case '\r':
    case '\t':
    case '\v':
        for (;;)
        {
            ac.push_back(uc);
            c = getch();
            if (c < 0)
            {
                debug_print("state = normal");
                state = state_normal;
                break;
            }
            uc = c;
            switch (uc)
            {
            case '\n':
                debug_print("state = normal");
                state = state_normal;
                // fall through...

            case ' ':
            case '\b':
            case '\f':
            case '\r':
            case '\t':
            case '\v':
                continue;
            }
            ungetch(uc);
            break;
        }
        value = ac.mkstr();
        return type_whitespace;

    case '$':
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
    case 'A': case 'B': case 'C': case 'D': case 'E':
    case 'F': case 'G': case 'H': case 'I': case 'J':
    case 'K': case 'L': case 'M': case 'N': case 'O':
    case 'P': case 'Q': case 'R': case 'S': case 'T':
    case 'U': case 'V': case 'W': case 'X': case 'Y':
    case 'Z':
    case '_':
    case 'a': case 'b': case 'c': case 'd': case 'e':
    case 'f': case 'g': case 'h': case 'i': case 'j':
    case 'k': case 'l': case 'm': case 'n': case 'o':
    case 'p': case 'q': case 'r': case 's': case 't':
    case 'u': case 'v': case 'w': case 'x': case 'y':
    case 'z':
        for (;;)
        {
            ac.push_back(uc);

            //
            // Get the next character
            //
            c = getch();
            if (c < 0)
            {
                debug_print("state = normal");
                state = state_normal;
                break;
            }
            uc = c;
            switch (uc)
            {
            case '$':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y':
            case 'Z':
            case '_':
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case 'w': case 'x': case 'y':
            case 'z':
                continue;

            default:
                ungetch(uc);
                break;
            }
            break;
        }
        value = ac.mkstr();
        break;

    case '"':
    case '\'':
        if (uc == terminator)
        {
            ac.push_back(uc);
            value = ac.mkstr();
            debug_print("state = normal");
            state = state_normal;
            break;
        }
        // Fall through...

    default:
        for (;;)
        {
            ac.push_back(uc);
            if (uc == '\\')
            {
                c = getch();
                if (c < 0)
                {
                    debug_print("state = normal");
                    state = state_normal;
                    break;
                }
                uc = c;
                ac.push_back(uc);
            }

            //
            // Get the next character
            //
            c = getch();
            if (c < 0)
            {
                debug_print("state = normal");
                state = state_normal;
                break;
            }
            uc = c;
            switch (uc)
            {
            case ' ':
            case '\b':
            case '\f':
            case '\n':
            case '\r':
            case '\t':
            case '\v':
            case '$':
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y':
            case 'Z':
            case '_':
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case 'w': case 'x': case 'y':
            case 'z':
                ungetch(uc);
                break;

            case '"':
            case '\'':
                if (uc == terminator)
                {
                    ungetch(uc);
                    break;
                }
                continue;

            default:
                continue;
            }
            break;
        }
        value = ac.mkstr();
        break;
    }
    return type_ink;
}
