//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009, 2010 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#include <lib/config.h>
#include <cstdio>
#include <libexplain/output.h>

#include <lib/fstrcmp.h>
#include <lib/sizeof.h>
#include <lib/source/c.h>
#include <lib/source/generic.h>
#include <lib/source/lines.h>


static source *
new_c(const rcstring &filename)
{
    return new source_c(filename);
}


static source *
new_generic(const rcstring &filename)
{
    return new source_generic(filename);
}


static source *
new_lines(const rcstring &filename)
{
    return new source_lines(filename);
}


struct table_t
{
    rcstring name;
    source *(*func)(const rcstring &filename);
    const char *description;
};

static const table_t table[] =
{
    {
        "generic",
        new_generic,
        "break the file into words, whitespace or punctuation"
    },
    {
        "c",
        new_c,
        "the file contains C source code, or similar",
    },
    {
        "lines",
        new_lines,
        "break the file up into lines"
    },
};


static bool
looks_like_c_code(const rcstring &filename)
{
    rcstring fn = filename.downcase();
    return
        (
            fn.ends_with(".c")
        ||
            fn.ends_with(".h")
        ||
            fn.ends_with(".cpp")
        ||
            fn.ends_with(".cc")
        );
}


source *
source::factory(const rcstring &filename, const rcstring &name)
{
    //
    // If no tokenizer name is specified, try to guess the name of the
    // tokenizer from the file name.
    //
    if (name.empty())
    {
        if (looks_like_c_code(filename))
            return new source_c(filename);
    }

    //
    // Default the tokenizer name.
    //
    rcstring lcname(name.empty() ? table[0].name : name.downcase());

    //
    // Search the table to find a factory function by name.
    //
    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (lcname == tp->name)
            return tp->func(filename);
    }

    //
    // Oops.
    // See if there is a similar name in the list.
    //
    rcstring closest;
    double closest_weight = 0.6;
    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        double w = fstrcmp(lcname.c_str(), tp->name.c_str());
        if (w > closest_weight)
        {
            closest = tp->name;
            closest_weight = w;
        }
    }
    if (!closest.empty())
    {
        explain_output_error_and_die
        (
            "tokenizer %s unknown, closest is %s",
            name.quote_c().c_str(),
            closest.quote_c().c_str()
        );
    }
    explain_output_error_and_die
    (
        "tokenizer %s unknown",
        name.quote_c().c_str()
    );
    // NOTREACHED
    return 0;
}


void
source::list()
{
    printf("Available tokenizers:\n");
    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        printf("%-15s %s", (tp->name + ":").c_str(), tp->description);
        if (tp == table)
            printf(" (default)");
        printf("\n");
    }
}
