//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#ifndef LIB_TOKEN_H
#define LIB_TOKEN_H

#include <lib/rcstring.h>

/**
  * The token class is used to represent an atomic ink+whitespace pair.
  */
class token
{
public:
    /**
      * The destructor.
      */
    virtual ~token();

    /**
      * The default constructor.
      */
    token();

    /**
      * The constructor.
      */
    token(const rcstring &white, const rcstring &ink);

    /**
      * The copy constructor.
      */
    token(const token &);

    /**
      * The assignment operator.
      */
    token &operator=(const token &);

    /**
      * The equality operator is used to compare two tokens for equality.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true if equal, false of not
      */
    bool
    operator==(const token &rhs)
        const
    {
        return (ink == rhs.ink);
    }

    /**
      * The inequality operator is used to compare two tokens for equality.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true if not equal, false of equal
      */
    bool
    operator!=(const token &rhs)
        const
    {
        return (ink != rhs.ink);
    }

    /**
      * The less than operator is used to compare two tokens.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true if less, false of greater than or equal
      */
    bool
    operator<(const token &rhs)
        const
    {
        return (ink < rhs.ink);
    }

    /**
      * The less than or eaual operator is used to compare two tokens.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true if less or equal, false of greater than
      */
    bool
    operator<=(const token &rhs)
        const
    {
        return (ink <= rhs.ink);
    }

    /**
      * The greater than operator is used to compare two tokens.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true if greater, false of less than or equal
      */
    bool
    operator>(const token &rhs)
        const
    {
        return (ink > rhs.ink);
    }

    /**
      * The greater than or equal operator is used to compare two tokens.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true if greater or equal, false of less than
      */
    bool
    operator>=(const token &rhs)
        const
    {
        return (ink >= rhs.ink);
    }

    /**
      * The print method is used to print the token (ink and white
      * space) onto the standard output.
      */
    void print() const;

    /**
      * The get_white method may be used to obtain the token's leading
      * white space.
      */
    rcstring get_white() const { return white; }

    /**
      * The get_ink method may be used to obtain the token's substantive
      * content.
      */
    rcstring get_ink() const { return ink; }

    /**
      * The representation method may be used to obtain a representation
      * of the token for debugging purposes.
      */
    rcstring representation() const;

private:
    /**
      * The white instance variable is used to remember the white spac
      * preceeding this token.
      */
    rcstring white;

    /**
      * The ink instance variable is used to remember the substatntive
      * content.
      */
    rcstring ink;
};

#endif // LIB_TOKEN_H
