//
//      contrast - compare files token by token
//      Copyright (C) 2006, 2009 Peter Miller
//
//      This program is free software; you can redistribute it and/or
//      modify it under the terms of the GNU General Public License,
//      version 3, as published by the Free Software Foundation.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program. If not, see
//      <http://www.gnu.org/licenses/>.
//

#ifndef LIB_TOKEN_LIST_H
#define LIB_TOKEN_LIST_H

#include <lib/token.h>

class source; // forward

/**
  * The token_list class is used to represent an ordered sequence of
  * tokens.
  */
class token_list
{
public:
    /**
      * The destructor.
      */
    virtual ~token_list();

    /**
      * The default constructor.
      */
    token_list();

    /**
      * The get method may be used to obtain a reference to the Nth
      * token in the list.
      *
      * @param n
      *     The number (zero based) of the token within the list.
      * @returns
      *     a constant reference to the token.
      */
    const token &get(size_t n) const { return item[n]; }

    /**
      * The array index operator method may be used to obtain a
      * reference to the Nth token in the list.
      *
      * @param n
      *     The number (zero based) of the token within the list.
      * @returns
      *     a constant reference to the token.
      */
    const token &operator[](size_t n) const { return get(n); }

    /**
      * The size method may be used to obtain the number of tokens in
      * the list.
      */
    size_t size() const { return length; }

    /**
      * The read method may be used to populate the token list by
      * reading a file and partitioning it into tokens.
      *
      * @param filename
      *     The name of the file to be read.
      * @param tokenizer_name
      *     The tokenizer to be used.  The empty string implies that the
      *     default tokenizer is to be used.
      */
    void read(const rcstring &filename, const rcstring &tokenizer_name = "");

    /**
      * The read method may be used to populate the token list by
      * reading from a token source.
      *
      * @param src
      *     The token source to read.
      */
    void read(source *src);

    /**
      * The print method may be used to print a representation of a
      * token.  This is usually for debugging purposes.
      */
    void print() const;

private:
    /**
      * The item instance variable is used to remember the address of a
      * dynamically allocated array of tokens.
      */
    token *item;

    /**
      * The length instance variable is used to remember how many
      * elements of the dynamically allocated array have been used to
      * date.
      */
    size_t length;

    /**
      * The maximum instance variable is used to remember the number of
      * elements in the dynamically allocated array of tokens.
      */
    size_t maximum;

    /**
      * The push_back method may be used to append another token to the
      * list.
      */
    void push_back(const rcstring &ink, const rcstring &white);

    /**
      * The copy constructor.
      */
    token_list(const token_list &);

    /**
      * The assignment operator.
      */
    token_list &operator=(const token_list &);
};

#endif // LIB_TOKEN_LIST_H
